﻿using System;
using System.Collections.Generic;
using System.Reflection;
using System.IO;
using System.Linq.Expressions;

namespace Understanding_Linq_3a
{
    public class CsvFileReader<TData> : IEnumerable<TData>, IDisposable where TData : new()
    {
        public CsvFileReader(string filename)
        {
            m_filename = filename;
            InitColumnMappings();
        }

        /// <summary>
        /// Analyze properties of TData and look for the CsvColumnAttribute
        /// </summary>
        private void InitColumnMappings()
        {
            m_columnMappings = new PropertyInfo[0];

            foreach (PropertyInfo propInfo in typeof(TData).GetProperties(BindingFlags.Instance | BindingFlags.Public))
            {
                CsvColumnAttribute[] columnInfo = (CsvColumnAttribute[])propInfo.GetCustomAttributes(typeof(CsvColumnAttribute), true);
                if ((columnInfo != null) && (columnInfo.Length == 1))
                {
                    int targetIndex = columnInfo[0].ColumnIndex;
                    if (m_columnMappings.Length <= targetIndex) Array.Resize(ref m_columnMappings, targetIndex + 1);
                    m_columnMappings[targetIndex] = propInfo;
                }
            }
        }

        /// <summary>
        /// Fill an empty TData from a line of a csv file.
        /// </summary>
        /// <param name="line"></param>
        /// <param name="data"></param>
        private void FillObject(string line, TData data)
        {
            //split columns
            string[] parts = line.Split(';');

            //load new object
            for (int idx = 0; idx < m_columnMappings.Length; idx++)
            {
                PropertyInfo targetProperty = m_columnMappings[idx];
                if (targetProperty != null)
                {
                    targetProperty.SetValue(data, Convert.ChangeType(parts[idx], targetProperty.PropertyType), null);
                }
            }
        }

        /// <summary>
        /// Array of the properties to load from file in the same order as the columns they are to be filled from.
        /// </summary>
        PropertyInfo[] m_columnMappings;
        string m_filename;

        #region IEnumerable<TData> Members

        public IEnumerator<TData> GetEnumerator()
        {
            return GetRemote(new RemoteExecutor()).GetEnumerator();
        }

        #endregion

        #region IEnumerable Members

     
        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        #endregion

        internal IEnumerable<TData> GetRemote(RemoteExecutor executor)
        {
            executor.Filename = m_filename;

            if (executor.ColumnsToLoad == null)
            {
                executor.ColumnsToLoad = ExtractMappedColumns();
            }

            string[][] rawdata = executor.GetData();

            int[] loadedColumns = Array.ConvertAll(executor.ColumnsToLoad.Split(','), s => int.Parse(s));

            //build TData objects from rawdata
            List<TData> result = new List<TData>();
            foreach (string[] row in rawdata)
            {
                TData data = new TData();

                for (int idx = 0; idx < m_columnMappings.Length; idx++)
                {
                    //only if the column was loaded 
                    if (Array.IndexOf(loadedColumns, idx) >= 0)
                    {
                        PropertyInfo targetProperty = m_columnMappings[idx];
                        if (targetProperty != null)
                        {
                            targetProperty.SetValue(data, Convert.ChangeType(row[idx], targetProperty.PropertyType), null);
                        }
                    }
                }
                result.Add(data);
            }
            return result.ToArray();

        }

        private string ExtractMappedColumns()
        {
            //load all columns mapped
            List<int> columns = new List<int>();
            for (int idx = 0; idx < m_columnMappings.Length; idx++)
            {
                if (m_columnMappings[idx] != null) columns.Add(idx);
            }
            return columns.ToArray().ToCsvString(",");
        }

        #region IDisposable Members

        public void Dispose()
        {
        }

        #endregion

        public Filter<TData> Where(Expression<Func<TData, bool>> filter)
        {
            return new Filter<TData>(this, filter);
        }
    }
}
